#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Defines the version number and details of qubovert."""

__all__ = (
    '__version__', '__author__', '__authoremail__', '__license__',
    '__sourceurl__', '__docsurl__', '__description__'
)


__version__ = "1.1.5"
__author__ = "Joseph T. Iosue"
__authoremail__ = "joe.iosue@yahoo.com"
__license__ = "Apache Software License 2.0"
__sourceurl__ = "https://github.com/jiosue/qubovert"
__docsurl__ = "https://qubovert.readthedocs.io"
__description__ = (
    "A package for formulating, simulating, and solving "
    "problems in boolean and spin form"
)
